
/*
 * bltGrAxis.h --
 *
 * Copyright 1991-1998 Lucent Technologies, Inc.
 *
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for any purpose and without fee is hereby
 * granted, provided that the above copyright notice appear in all
 * copies and that both that the copyright notice and warranty
 * disclaimer appear in supporting documentation, and that the names
 * of Lucent Technologies any of their entities not be used in
 * advertising or publicity pertaining to distribution of the software
 * without specific, written prior permission.
 *
 * Lucent Technologies disclaims all warranties with regard to this
 * software, including all implied warranties of merchantability and
 * fitness.  In no event shall Lucent Technologies be liable for any
 * special, indirect or consequential damages or any damages
 * whatsoever resulting from loss of use, data or profits, whether in
 * an action of contract, negligence or other tortuous action, arising
 * out of or in connection with the use or performance of this
 * software.
 */

#ifndef _BLT_GR_AXIS_H
#define _BLT_GR_AXIS_H

#include "bltList.h"

/*
 * -------------------------------------------------------------------
 *
 * Range --
 *
 *	Designates a range of values by a minimum and maximum limit.
 *
 * -------------------------------------------------------------------
 */
typedef struct {
    double min, max, range;
} Range;

#define SetLimits(lim, lo, hi) \
	((lim).min = (lo), (lim).max = (hi), (lim).range = ((hi) - (lo)))
#define SetRange(lim) \
	((lim).range = ((lim).max > (lim).min) ? ((lim).max - (lim).min) : 0.0)

/*
 * ----------------------------------------------------------------------
 *
 * TickLabel --
 *
 * 	Structure containing the X-Y screen coordinates of the tick
 * 	label (anchored at its center).
 *
 * ----------------------------------------------------------------------
 */
typedef struct TickLabel {
    short int x, y;
} TickLabel;

/*
 * ----------------------------------------------------------------------
 *
 * Ticks --
 *
 * 	Structure containing information where the ticks (major or
 *	minor) will be displayed on the graph.
 *
 * ----------------------------------------------------------------------
 */
typedef struct {
    int nTicks;			/* # of ticks on axis */
    double tickArr[1];		/* Array of tick values (malloc-ed). */
} Ticks;

/*
 * ----------------------------------------------------------------------
 *
 * TickSweep --
 *
 * 	Structure containing information where the ticks (major or
 *	minor) will be displayed on the graph.
 *
 * ----------------------------------------------------------------------
 */
typedef struct {
    double initial;		/* Initial value */
    double step;		/* Size of interval */
    int nSteps;			/* Number of intervals. */
} TickSweep;

/*
 * ----------------------------------------------------------------------
 *
 * Axis --
 *
 * 	Structure contains options controlling how the axis will be
 * 	displayed.
 *
 * ----------------------------------------------------------------------
 */
typedef struct {
    char *name;			/* Axis identifier */
    int type;			/* Type of axis: AXIS_TYPE_X or AXIS_TYPE_Y */
    int deletePending;		/* Indicates that the axis was
				 * scheduled for deletion. The actual
				 * deletion may be deferred until the
				 * axis is no longer in use.  */

    int refCount;		/* Number of elements referencing this
				 * axis. */

    Tcl_HashEntry *hashPtr;	/* Points to axis entry in hash
				 * table. Used to quickly remove axis
				 * entries. */

    int logScale;		/* If non-zero, scale the axis values
				 * logarithmically. */

    int hidden;			/* If non-zero, don't display the
				 * axis title, ticks, or line. */

    int showTicks;		/* If non-zero, display tick marks and
				 * labels. */

    int descending;		/* If non-zero, display the range of
				 * values on the axis in descending
				 * order, from high to low. */

    int looseMin, looseMax;	/* If non-zero, axis range extends to
				 * the outer major ticks, otherwise at
				 * the limits of the data values. This
				 * is overriddened by setting the -min
				 * and -max options.  */

    unsigned int flags;		/* Set bit field definitions below */

    /*
     * AXIS_AUTO_MIN		Set axis minimum based upon data points
     * AXIS_AUTO_MAX		Set axis maximum based upon data points
     * AXIS_MAPS_ELEM		Axis is currently mapping element coordinates
     * AXIS_DRAWN		Axis is designated as a logical axis
     * AXIS_CONFIG_DIRTY
     * AXIS_TYPE_X		Axis is used as a X-axis
     * AXIS_TYPE_Y		Axis is used as a Y-axis
     */

    char *titleText;		/* Title of axis. */
    TextStyle titleStyle;	/* Text attributes (color, font,
				 * rotation, etc.)  of the axis
				 * title. */

    short int titleX, titleY;	/* Position of the title */
    short int titleWidth, titleHeight;	

    int lineWidth;		/* Width of lines representing axis
				 * (including ticks).  If zero, then
				 * no axis lines or ticks are
				 * drawn. */

    char **limitsFormats;	/* One or two strings of sprintf-like
				 * formats describing how to display
				 * virtual axis limits. If NULL,
				 * display no limits. */
    int nFormats;

    TextStyle limitsStyle;	/* Text attributes (color, font,
				 * rotation, etc.)  of the limits. */

    double autoRange;		/* Size of a sliding window of values
				 * used to scale the axis automatically
				 * as new data values are added. The axis
				 * will always display the latest values
				 * in this range. */

    double shiftBy;		/* Shift maximum by this interval. */

    int tickLength;		/* Length of major ticks in pixels */

    TextStyle tickStyle;	/* Text attributes (color, font, rotation, 
				 * etc.) for labels at each major tick. */

    char *formatCmd;		/* Specifies a Tcl command, to be invoked
				 * by the axis whenever it has to generate 
				 * tick labels. */

    char *scrollCmdPrefix;
    int scrollUnits;

    double min, max;		/* If set, overrides the computed range of 
				 * the axis using the -min and -max options. 
				 * Otherwise contains values determined by 
				 * auto-scaling. */

    Range dataRange;		/* Range of data values of elements mapped 
				 * to this axis. This is used to auto-scale 
				 * the axis in "tight" mode. */

    Range tickRange;		/* Smallest and largest major tick values
				 * for the axis.  The tick values lie outside
				 * the range of data values.  This is used to
				 * auto-scale the axis in "loose" mode. */

    double prevMin, prevMax;

    double reqStep;		/* If > 0.0, overrides the computed major 
				 * tick interval.  Otherwise a stepsize 
				 * is automatically calculated, based 
				 * upon the range of elements mapped to the 
				 * axis. The default value is 0.0. */

    GC tickGC;			/* Graphics context for axis and tick labels */

    Ticks *reqMajorPtr;		/* Requested major tick positions. If
				 * set by the user (non-NULL), points to
				 * a malloc-ed array of values representing
				 * each major tick value. */

    Ticks *reqMinorPtr;		/* Requested minor tick positions. If
				 * set by the user (non-NULL), points
				 * to a malloc-ed array of values
				 * representing each major tick
				 * value. */

    int reqNumMinorTicks;	/* If non-zero, represents the
				 * requested the number of minor ticks
				 * to be uniformally displayed along
				 * each major tick. */

    TickSweep minorTicks, majorTicks;


    /* The following fields are specific to logical axes */

    Blt_ChainLink *linkPtr;	/* Axis link in margin list. */
    Blt_Chain *chainPtr;

    short int width, height;	/* Extents of axis */

    Ticks *genMajorPtr;		/* Generated major tick positions */
    Ticks *genMinorPtr;		/* Generated minor tick positions */

    Ticks *majorPtr;		/* Major tick positions */
    Ticks *minorPtr;		/* Minor tick positions */

    XSegment *segArr;		/* Array of line segments representing
				 * the major and minor ticks, but also
				 * the axis line itself. The segment
				 * coordinates are relative to the
				 * axis. */

    int nSegments;		/* Number of segments in the above array. */

    Blt_List labelList;		/* Contains major tick label strings 
				 * and their offsets along the axis. */

} Axis;

#define AXIS_TYPE_NONE	0
#define AXIS_TYPE_X	(1<<0)
#define AXIS_TYPE_Y	(1<<1)
#define AXIS_TYPE_MASK  (AXIS_TYPE_X | AXIS_TYPE_Y)

#define AXIS_AUTO_MAX	(1<<2)	/* Set axis maximum based upon data points */
#define AXIS_AUTO_MIN	(1<<3)	/* Set axis minimum based upon data points */

#define AXIS_MAPS_ELEM	(1<<4)	/* Axis is used to map element coordinates */
#define AXIS_CONFIG_DIRTY (1<<5)

#define AXIS_ALLOW_NULL (1<<12)

/*
 * -------------------------------------------------------------------
 *
 * Axis2D --
 *
 *	The pair of axes mapping a point onto the graph.
 *
 * -------------------------------------------------------------------
 */
typedef struct {
    Axis *x, *y;
} Axis2D;

#endif /* _BLT_GR_AXIS_H */
